from pathlib import Path
from shutil import rmtree

from belt_nlp.bert_with_pooling import BertClassifierWithPooling

MODEL_PARAMS = {
    "batch_size": 1,
    "learning_rate": 5e-5,
    "epochs": 1,
    "chunk_size": 510,
    "stride": 256,
    "minimal_chunk_length": 1,
    "pooling_strategy": "mean",
}


def test_fit_and_predict():
    """The test is quite naive, but it goes through all the methods."""
    params = MODEL_PARAMS
    model = BertClassifierWithPooling(**params, device="cpu")
    x_train = ["carrot", "cucumber", "tomato", "potato"]
    y_train = [True] * len(x_train)

    x_test = ["pepper", "eggplant"]
    expected_classes = [True, True]

    model.fit(x_train, y_train)
    predictions = model.predict(x_test)

    assert len(predictions) == 2
    assert [x[0] for x in predictions] == expected_classes
    for cls, score in predictions:
        assert isinstance(cls, bool)
        assert isinstance(score, float)


def test_prediction_order():
    """Check if the order of predictions is preserved."""
    params = MODEL_PARAMS
    model = BertClassifierWithPooling(**params, device="cpu")
    x_train = ["carrot", "cucumber", "tomato", "potato"]
    y_train = [True] * len(x_train)

    x_test = ["pepper"] * 99 + ["chair"] * 1

    model.fit(x_train, y_train)
    predicted_scores = model.predict_scores(x_test)

    expected_score_pepper = model.predict_scores(["pepper"])[0]
    expected_score_chair = model.predict_scores(["chair"])[0]

    # Test if only last prediction is different
    # This will fail if the predictions are shuffled
    for i in range(len(x_test) - 1):
        assert predicted_scores[i] == expected_score_pepper
    assert predicted_scores[-1] == expected_score_chair


def test_save_and_load():
    params = MODEL_PARAMS
    model = BertClassifierWithPooling(**params, device="cpu")
    path = Path("tmp_bert_model_test_dir")

    model.save(str(path))

    try:
        model_loaded = BertClassifierWithPooling.load(str(path), device="cpu")
        # assert types to be more specific than 'isinstance()'
        assert type(model_loaded.tokenizer) == type(model.tokenizer)  # noqa: E721
        assert type(model_loaded.neural_network) == type(model.neural_network)  # noqa: E721
    finally:
        rmtree(path)



import pytest
from torch import Tensor

from belt_nlp.exceptions import InconsistentSplittingParamsException
from belt_nlp.splitting import split_overlapping
from belt_nlp.tensor_utils import list_of_tensors_deep_equal

EXAMPLE_TENSOR = Tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])


def test_split_overlapping():
    expected_result = [Tensor([1, 2, 3, 4, 5]), Tensor([6, 7, 8, 9, 10])]
    splitted = split_overlapping(EXAMPLE_TENSOR, chunk_size=5, stride=5, minimal_chunk_length=5)
    assert list_of_tensors_deep_equal(splitted, expected_result)

    expected_result = [
        Tensor([1, 2, 3, 4, 5]),
        Tensor([2, 3, 4, 5, 6]),
        Tensor([3, 4, 5, 6, 7]),
        Tensor([4, 5, 6, 7, 8]),
        Tensor([5, 6, 7, 8, 9]),
        Tensor([6, 7, 8, 9, 10]),
    ]
    splitted = split_overlapping(EXAMPLE_TENSOR, chunk_size=5, stride=1, minimal_chunk_length=5)
    assert list_of_tensors_deep_equal(splitted, expected_result)

    expected_result = [
        Tensor([1, 2, 3, 4, 5]),
        Tensor([2, 3, 4, 5, 6]),
        Tensor([3, 4, 5, 6, 7]),
        Tensor([4, 5, 6, 7, 8]),
        Tensor([5, 6, 7, 8, 9]),
        Tensor([6, 7, 8, 9, 10]),
        Tensor([7, 8, 9, 10]),
        Tensor([8, 9, 10]),
    ]
    splitted = split_overlapping(EXAMPLE_TENSOR, chunk_size=5, stride=1, minimal_chunk_length=3)
    assert list_of_tensors_deep_equal(splitted, expected_result)

    expected_result = [
        Tensor([1, 2, 3, 4, 5, 6, 7, 8, 9]),
        Tensor([2, 3, 4, 5, 6, 7, 8, 9, 10]),
        Tensor([3, 4, 5, 6, 7, 8, 9, 10]),
        Tensor([4, 5, 6, 7, 8, 9, 10]),
        Tensor([5, 6, 7, 8, 9, 10]),
        Tensor([6, 7, 8, 9, 10]),
        Tensor([7, 8, 9, 10]),
        Tensor([8, 9, 10]),
    ]
    splitted = split_overlapping(EXAMPLE_TENSOR, chunk_size=9, stride=1, minimal_chunk_length=3)
    assert list_of_tensors_deep_equal(splitted, expected_result)


def test_too_large_chunk_exception():
    with pytest.raises(InconsistentSplittingParamsException):
        split_overlapping(EXAMPLE_TENSOR, chunk_size=511, stride=2, minimal_chunk_length=1)


def test_stride_larger_than_chunk_size_exception():
    with pytest.raises(InconsistentSplittingParamsException):
        split_overlapping(EXAMPLE_TENSOR, chunk_size=3, stride=4, minimal_chunk_length=1)


def test_minimal_length_larger_than_chunk_size_exception():
    with pytest.raises(InconsistentSplittingParamsException):
        split_overlapping(EXAMPLE_TENSOR, chunk_size=3, stride=2, minimal_chunk_length=4)



import pandas as pd

from belt_nlp.bert_with_pooling import BertClassifierWithPooling
from belt_nlp.tensor_utils import list_of_tensors_deep_equal

MODEL_PARAMS = {
    "batch_size": 1,
    "learning_rate": 5e-5,
    "epochs": 1,
    "chunk_size": 510,
    "stride": 510,
    "minimal_chunk_length": 510,
    "pooling_strategy": "mean",
}
MODEL = BertClassifierWithPooling(**MODEL_PARAMS, device="cpu")

SAMPLE_DATA_PATH = "sample_data/sample_data_eng.csv"
SHORT_REVIEW = pd.read_csv(SAMPLE_DATA_PATH).loc[0, "sentence"]
LONG_REVIEW = pd.read_csv(SAMPLE_DATA_PATH).loc[961, "sentence"]


def test_tokenize():
    tokens_short = MODEL._tokenize([SHORT_REVIEW])
    assert list(tokens_short["input_ids"][0].shape) == [1, 512]
    assert list(tokens_short["attention_mask"][0].shape) == [1, 512]

    tokens_long = MODEL._tokenize([LONG_REVIEW])
    expected_number_of_chunks = 3
    assert list(tokens_long["input_ids"][0].shape) == [expected_number_of_chunks, 512]
    assert list(tokens_long["attention_mask"][0].shape) == [expected_number_of_chunks, 512]

    tokens_two_texts_in_batch = MODEL._tokenize([SHORT_REVIEW, LONG_REVIEW])
    expected_input_ids = tokens_short["input_ids"] + tokens_long["input_ids"]
    expected_attention_mask = tokens_short["attention_mask"] + tokens_long["attention_mask"]
    assert list_of_tensors_deep_equal(tokens_two_texts_in_batch["input_ids"], expected_input_ids)
    assert list_of_tensors_deep_equal(tokens_two_texts_in_batch["attention_mask"], expected_attention_mask)



from __future__ import annotations
from abc import ABC, abstractmethod
import json
from pathlib import Path
from typing import Any, Optional, Union

import torch
from torch import Tensor
from torch.nn import BCELoss, DataParallel, Module, Linear, Sigmoid
from torch.optim import AdamW, Optimizer
from torch.utils.data import Dataset, RandomSampler, SequentialSampler, DataLoader
from transformers import AutoModel, AutoTokenizer, BatchEncoding, BertModel, PreTrainedTokenizerBase, RobertaModel


class BertClassifier(ABC):
    """
    The "device" parameter can have the following values:
        - "cpu" - The model will be loaded on CPU.
        - "cuda" - The model will be loaded on single GPU.
        - "cuda:i" - The model will be loaded on the specific single GPU with the index i.

    It is also possible to use multiple GPUs. In order to do this:
        - Set device to "cuda".
        - Set many_gpu flag to True.
        - As default it will use all of them.

    To use only selected GPUs - set the environmental variable CUDA_VISIBLE_DEVICES.
    """

    @abstractmethod
    def __init__(
        self,
        batch_size: int,
        learning_rate: float,
        epochs: int,
        accumulation_steps: int = 1,
        tokenizer: Optional[PreTrainedTokenizerBase] = None,
        neural_network: Optional[Module] = None,
        pretrained_model_name_or_path: Optional[str] = "bert-base-uncased",
        device: str = "cuda:0",
        many_gpus: bool = False,
    ):
        if not tokenizer:
            tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name_or_path)
        if not neural_network:
            bert = AutoModel.from_pretrained(pretrained_model_name_or_path)
            neural_network = BertClassifierNN(bert)

        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.accumulation_steps = accumulation_steps
        self._params = {
            "batch_size": self.batch_size,
            "learning_rate": self.learning_rate,
            "epochs": self.epochs,
            "accumulation_steps": self.accumulation_steps
        }
        self.device = device
        self.many_gpus = many_gpus
        self.tokenizer = tokenizer
        self.neural_network = neural_network
        self.collate_fn = None

        self.neural_network.to(device)
        if device.startswith("cuda") and many_gpus:
            self.neural_network = DataParallel(self.neural_network)

    def fit(self, x_train: list[str], y_train: list[bool], epochs: Optional[int] = None) -> None:
        if not epochs:
            epochs = self.epochs
        optimizer = AdamW(self.neural_network.parameters(), lr=self.learning_rate)

        tokens = self._tokenize(x_train)
        dataset = TokenizedDataset(tokens, y_train)
        dataloader = DataLoader(
            dataset, sampler=RandomSampler(dataset), batch_size=self.batch_size, collate_fn=self.collate_fn
        )
        for epoch in range(epochs):
            self._train_single_epoch(dataloader, optimizer)

    def predict(self, x: list[str], batch_size: Optional[int] = None) -> list[tuple[bool, float]]:
        if not batch_size:
            batch_size = self.batch_size
        scores = self.predict_scores(x, batch_size)
        classes = [i >= 0.5 for i in scores]
        return list(zip(classes, scores))

    def predict_classes(self, x: list[str], batch_size: Optional[int] = None) -> list[bool]:
        if not batch_size:
            batch_size = self.batch_size
        scores = self.predict_scores(x, batch_size)
        classes = [i >= 0.5 for i in scores]
        return classes

    def predict_scores(self, x: list[str], batch_size: Optional[int] = None) -> list[float]:
        if not batch_size:
            batch_size = self.batch_size
        tokens = self._tokenize(x)
        dataset = TokenizedDataset(tokens)
        dataloader = DataLoader(
            dataset, sampler=SequentialSampler(dataset), batch_size=batch_size, collate_fn=self.collate_fn
        )
        total_predictions = []

        # deactivate dropout layers
        self.neural_network.eval()
        for step, batch in enumerate(dataloader):
            # deactivate autograd
            with torch.no_grad():
                predictions = self._evaluate_single_batch(batch)
                total_predictions.extend(predictions.tolist())
        return total_predictions

    @abstractmethod
    def _tokenize(self, texts: list[str]) -> BatchEncoding:
        pass

    def _train_single_epoch(self, dataloader: DataLoader, optimizer: Optimizer) -> None:
        self.neural_network.train()
        cross_entropy = BCELoss()

        for step, batch in enumerate(dataloader):

            labels = batch[-1].float().cpu()
            predictions = self._evaluate_single_batch(batch)
            loss = cross_entropy(predictions, labels) / self.accumulation_steps
            loss.backward()

            if ((step + 1) % self.accumulation_steps == 0) or (step + 1 == len(dataloader)):
                optimizer.step()
                optimizer.zero_grad()

    @abstractmethod
    def _evaluate_single_batch(self, batch: tuple[Tensor]) -> Tensor:
        pass

    def save(self, model_dir: str) -> None:
        model_dir = Path(model_dir)
        model_dir.mkdir(parents=True, exist_ok=True)
        with open(file=model_dir / "params.json", mode="w", encoding="utf-8") as file:
            json.dump(self._params, file)
        self.tokenizer.save_pretrained(model_dir)
        if self.many_gpus:
            torch.save(self.neural_network.module, model_dir / "model.bin")
        else:
            torch.save(self.neural_network, model_dir / "model.bin")

    @classmethod
    def load(cls, model_dir: str, device: str = "cuda:0", many_gpus: bool = False) -> BertClassifier:
        model_dir = Path(model_dir)
        with open(file=model_dir / "params.json", mode="r", encoding="utf-8") as file:
            params = json.load(file)
        tokenizer = AutoTokenizer.from_pretrained(model_dir)
        neural_network = torch.load(f=model_dir / "model.bin", map_location=device)
        return cls(
            **params,
            tokenizer=tokenizer,
            neural_network=neural_network,
            pretrained_model_name_or_path=None,
            device=device,
            many_gpus=many_gpus,
        )


class BertClassifierNN(Module):
    def __init__(self, model: Union[BertModel, RobertaModel]):
        super().__init__()
        self.model = model

        # classification head
        self.linear = Linear(768, 1)
        self.sigmoid = Sigmoid()

    def forward(self, input_ids: Tensor, attention_mask: Tensor) -> Tensor:
        x = self.model(input_ids, attention_mask)
        x = x[0][:, 0, :]  # take <s> token (equiv. to [CLS])

        # classification head
        x = self.linear(x)
        x = self.sigmoid(x)
        return x


class TokenizedDataset(Dataset):
    """Dataset for tokens with optional labels."""

    def __init__(self, tokens: BatchEncoding, labels: Optional[list] = None):
        self.input_ids = tokens["input_ids"]
        self.attention_mask = tokens["attention_mask"]
        self.labels = labels

    def __len__(self) -> int:
        return len(self.input_ids)

    def __getitem__(self, idx: int) -> Union[tuple[Tensor, Tensor, Any], tuple[Tensor, Tensor]]:
        if self.labels:
            return self.input_ids[idx], self.attention_mask[idx], self.labels[idx]
        return self.input_ids[idx], self.attention_mask[idx]



from __future__ import annotations
from typing import Optional
import torch
from torch import Tensor
from torch.nn import Module
from transformers import BatchEncoding, PreTrainedTokenizerBase

from belt_nlp.bert import BertClassifier


class BertClassifierTruncated(BertClassifier):
    def __init__(
        self,
        batch_size: int,
        learning_rate: float,
        epochs: int,
        accumulation_steps: int = 1,
        tokenizer: Optional[PreTrainedTokenizerBase] = None,
        neural_network: Optional[Module] = None,
        pretrained_model_name_or_path: Optional[str] = "bert-base-uncased",
        device: str = "cuda:0",
        many_gpus: bool = False,
    ):
        super().__init__(
            batch_size,
            learning_rate,
            epochs,
            accumulation_steps,
            tokenizer,
            neural_network,
            pretrained_model_name_or_path,
            device,
            many_gpus,
        )

    def _tokenize(self, texts: list[str]) -> BatchEncoding:
        """
        Transforms list of N texts to the BatchEncoding, that is the dictionary with the following keys:
            - input_ids - Tensor of the size N x 512 of token ids.
                This is stacked Tensor of encodings of each text.
                Values of the tensor are integers.
            - attention_mask - Tensor of the size N x 512 of attention masks.
                This is stacked Tensor of encodings of each text.
                Values of the tensor are booleans.

        If the text is longer than 512 tokens - the rest of it is ignored.
        If the text is shorter than 512 tokens - it is padded to have exactly 512 tokens.
        """
        tokens = self.tokenizer(texts, max_length=512, padding=True, truncation=True, return_tensors="pt")
        return tokens

    def _evaluate_single_batch(self, batch: tuple[Tensor]) -> Tensor:
        batch = [t.to(self.device) for t in batch]
        model_input = batch[:2]

        predictions = self.neural_network(*model_input)
        predictions = torch.flatten(predictions).cpu()
        return predictions
